@@ -1,36 +1,42 @@
 package Apache2::AuthCookie;
+{
+  $Apache2::AuthCookie::VERSION = '3.19_01';
+}
+
+# ABSTRACT: Perl Authentication and Authorization via cookies
 
 use strict;
 
 use Carp;
 use CGI '3.12';
-use mod_perl2 '1.9922';
+use mod_perl2 '1.99022';
 
 use Apache::AuthCookie::Util;
+use Apache2::AuthCookie::Params;
 use Apache2::RequestRec;
 use Apache2::RequestUtil;
 use Apache2::Log;
 use Apache2::Access;
 use Apache2::Response;
 use Apache2::Util;
+use Apache::AuthCookie::Autobox;
 use APR::Table;
-use Apache2::Const qw(:common M_GET HTTP_FORBIDDEN HTTP_MOVED_TEMPORARILY);
-use vars qw($VERSION);
-
-# $Id: AuthCookie.pm 226 2008-03-03 17:37:04Z mschout $
-$VERSION = '3.12_01';
+use Apache2::Const qw(:common M_GET HTTP_FORBIDDEN HTTP_MOVED_TEMPORARILY HTTP_OK);
 
 sub recognize_user {
     my ($self, $r) = @_;
 
+    # only check if user is not already set
+    return DECLINED unless $r->user->is_blank;
+
     my $debug = $r->dir_config("AuthCookieDebug") || 0;
 
     my $auth_type = $r->auth_type;
     my $auth_name = $r->auth_name;
 
-    return DECLINED unless $auth_type and $auth_name;
+    return DECLINED if $auth_type->is_blank or $auth_name->is_blank;
 
-    return DECLINED unless $r->headers_in->get('Cookie');
+    return DECLINED if $r->headers_in->get('Cookie')->is_blank;
 
     my $cookie = $self->key($r);
     my $cookie_name = $self->cookie_name($r);
@@ -38,11 +44,11 @@ sub recognize_user {
     $r->server->log_error("cookie $cookie_name is $cookie")
         if $debug >= 2;
 
-    return DECLINED unless $cookie;
+    return DECLINED if $cookie->is_blank;
 
     my ($user,@args) = $auth_type->authen_ses_key($r, $cookie);
 
-    if ($user and scalar @args == 0) {
+    if (!$user->is_blank and scalar @args == 0) {
         $r->server->log_error("user is $user") if $debug >= 2;
 
         # send cookie with update expires timestamp if session timeout is on
@@ -56,7 +62,7 @@ sub recognize_user {
         return $auth_type->custom_errors($r, $user, @args);
     }
 
-    return OK;
+    return $user->is_blank ? DECLINED : OK;
 }
 
 sub cookie_name {
@@ -99,14 +105,12 @@ sub remove_cookie {
     );
 
     $r->err_headers_out->add("Set-Cookie" => "$str");
-    $r->server->log_error("removed_cookie ".
-                          $r->err_headers_out->get("Set-Cookie"))
-        if $debug >= 2;
+    $r->server->log_error("removed cookie $cookie_name") if $debug >= 2;
 }
 
 # convert current request to GET
 sub _convert_to_get {
-    my ($self, $r, $args) = @_;
+    my ($self, $r) = @_;
 
     return unless $r->method eq 'POST';
 
@@ -114,14 +118,16 @@ sub _convert_to_get {
 
     $r->server->log_error("Converting POST -> GET") if $debug >= 2;
 
-    my @pairs =();
-    while (my ($name, $value) = each %$args) {
+    my $args = $self->params($r);
+
+    my @pairs = ();
+
+    for my $name ($args->param) {
         # we dont want to copy login data, only extra data
         next if $name eq 'destination'
              or $name =~ /^credential_\d+$/;
 
-        $value = '' unless defined $value;
-        for my $v (split /\0/, $value) {
+        for my $v ($args->param($name)) {
             push @pairs, escape_uri($r, $name) . '=' . escape_uri($r, $v);
         }
     }
@@ -139,14 +145,10 @@ sub escape_uri {
 }
 
 # get GET or POST data and return hash containing the data.
-sub _get_form_data {
+sub params {
     my ($self, $r) = @_;
 
-    my $data = '';
-
-    my $cgi = CGI->new($r);
-
-    return $cgi->Vars();
+    return Apache2::AuthCookie::Params->new($r);
 }
 
 sub login {
@@ -157,13 +159,13 @@ sub login {
     my $auth_type = $r->auth_type;
     my $auth_name = $r->auth_name;
 
-    my %args = $self->_get_form_data($r);
+    my $params = $self->params($r);
 
     if ($r->method eq 'POST') {
-        $self->_convert_to_get($r, \%args);
+        $self->_convert_to_get($r);
     }
 
-    unless (exists $args{'destination'}) {
+    unless (defined $params->param('destination')) {
         $r->server->log_error("No key 'destination' found in form data");
         $r->subprocess_env('AuthCookieReason', 'no_cookie');
         return $auth_type->login_form($r);
@@ -171,10 +173,11 @@ sub login {
 
     # Get the credentials from the data posted by the client
     my @credentials;
-    for (my $i = 0; exists $args{"credential_$i"}; $i++) {
+    for (my $i = 0; defined $params->param("credential_$i"); $i++) {
         my $key = "credential_$i";
-        $r->server->log_error("$key $args{$key}") if $debug >= 2;
-        push @credentials, $args{$key};
+        my $val = $params->param($key);
+        $r->server->log_error("$key $val") if $debug >= 2;
+        push @credentials, $val;
     }
 
     # save creds in pnotes so login form script can use them if it wants to
@@ -185,7 +188,7 @@ sub login {
     unless ($ses_key) {
         $r->server->log_error("Bad credentials") if $debug >= 2;
         $r->subprocess_env('AuthCookieReason', 'bad_credentials');
-        $r->uri($args{'destination'});
+        $r->uri($params->param('destination'));
         return $auth_type->login_form($r);
     }
 
@@ -199,11 +202,11 @@ sub login {
     $self->handle_cache($r);
 
     if ($debug >= 2) {
-        $r->server->log_error("redirect to $args{destination}");
+        $r->server->log_error("redirect to ", $params->param('destination'));
     }
 
     $r->headers_out->set(
-        "Location" => $self->untaint_destination($args{'destination'}));
+        "Location" => $self->untaint_destination($params->param('destination')));
 
     return HTTP_MOVED_TEMPORARILY;
 }
@@ -269,7 +272,7 @@ sub authenticate {
     if ($ses_key_cookie) {
         my ($auth_user, @args) = $auth_type->authen_ses_key($r, $ses_key_cookie);
 
-        if ($auth_user and scalar @args == 0) {
+        if (!$auth_user->is_blank and scalar @args == 0) {
             # We have a valid session key, so we return with an OK value.
             # Tell the rest of Apache what the authentication method and
             # user is.
@@ -312,10 +315,8 @@ sub login_form {
 
     my $auth_name = $r->auth_name;
 
-    my %args = $self->_get_form_data($r);
-
     if ($r->method eq 'POST') {
-        $self->_convert_to_get($r, \%args);
+        $self->_convert_to_get($r);
     }
 
     # There should be a PerlSetVar directive that gives us the URI of
@@ -327,9 +328,26 @@ sub login_form {
         return SERVER_ERROR;
     }
 
-    $r->custom_response(HTTP_FORBIDDEN, $authen_script);
+    my $status = $self->login_form_status($r);
+    $status = HTTP_FORBIDDEN unless defined $status;
 
-    return HTTP_FORBIDDEN;
+    $r->custom_response($status, $authen_script);
+
+    return $status;
+}
+
+sub login_form_status {
+    my ($self, $r) = @_;
+
+    my $ua = $r->headers_in->get('User-Agent')
+        or return HTTP_FORBIDDEN;
+
+    if (Apache::AuthCookie::Util::understands_forbidden_response($ua)) {
+        return HTTP_FORBIDDEN;
+    }
+    else {
+        return HTTP_OK;
+    }
 }
 
 sub satisfy_is_valid {
@@ -376,8 +394,8 @@ sub authorize {
 
     $r->server->log_error("authorize user=$user type=$auth_type") if $debug >=3;
 
-    unless ($user) {
-        # user is either undef or =0 which means the authentication failed
+    if ($user->is_blank) {
+        # the authentication failed
         $r->server->log_error("No user authenticated", $r->uri);
         return HTTP_FORBIDDEN;
     }
@@ -440,16 +458,21 @@ sub send_cookie {
         %$cookie_args
     );
 
-    my $auth_name = $r->auth_name;
-
-    # add a P3P header if user has configured it.
-    if (my $p3p = $r->dir_config("${auth_name}P3P")) {
-        $r->err_headers_out->set(P3P => $p3p);
-    }
+    $self->send_p3p($r);
 
     $r->err_headers_out->add("Set-Cookie" => $cookie);
 }
 
+# send P3P header if configured with ${auth_name}P3P
+sub send_p3p {
+    my ($self, $r) = @_;
+
+    my $auth_name = $r->auth_name;
+
+    if (my $p3p = $r->dir_config("${auth_name}P3P")) {
+        $r->err_headers_out->set(P3P => $p3p);
+    }
+}
 
 # cookie_string takes named parameters:
 #    request
@@ -517,19 +540,23 @@ sub get_cookie_path {
 
 1;
 
-__END__
+=pod
 
 =head1 NAME
 
 Apache2::AuthCookie - Perl Authentication and Authorization via cookies
 
+=head1 VERSION
+
+version 3.19_01
+
 =head1 SYNOPSIS
 
 Make sure your mod_perl is at least 1.24, with StackedHandlers,
 MethodHandlers, Authen, and Authz compiled in.
 
  # In httpd.conf or .htaccess:
- PerlModule Sample::AuthCookieHandler
+ PerlModule Sample::Apache2::AuthCookieHandler
  PerlSetVar WhatEverPath /
  PerlSetVar WhatEverLoginScript /login.pl
 
@@ -571,38 +598,34 @@ MethodHandlers, Authen, and Authz compiled in.
 
  # These documents require user to be logged in.
  <Location /protected>
-  AuthType Sample::AuthCookieHandler
+  AuthType Sample::Apache2::AuthCookieHandler
   AuthName WhatEver
-  PerlAuthenHandler Sample::AuthCookieHandler->authenticate
-  PerlAuthzHandler Sample::AuthCookieHandler->authorize
+  PerlAuthenHandler Sample::Apache2::AuthCookieHandler->authenticate
+  PerlAuthzHandler Sample::Apache2::AuthCookieHandler->authorize
   require valid-user
  </Location>
 
  # These documents don't require logging in, but allow it.
  <FilesMatch "\.ok$">
-  AuthType Sample::AuthCookieHandler
+  AuthType Sample::Apache2::AuthCookieHandler
   AuthName WhatEver
-  PerlFixupHandler Sample::AuthCookieHandler->recognize_user
+  PerlFixupHandler Sample::Apache2::AuthCookieHandler->recognize_user
  </FilesMatch>
 
  # This is the action of the login.pl script above.
  <Files LOGIN>
-  AuthType Sample::AuthCookieHandler
+  AuthType Sample::Apache2::AuthCookieHandler
   AuthName WhatEver
   SetHandler perl-script
-  PerlResponseHandler Sample::AuthCookieHandler->login
+  PerlResponseHandler Sample::Apache2::AuthCookieHandler->login
  </Files>
 
 =head1 DESCRIPTION
 
-WARNING! This is an early version of B<Apache::AuthCookie> for mod_perl version
-2.  If you are running under mod_perl version 2, this module should be
-considered of beta quality.  If you are a subclass author and would like to see
-additional changes (besides those listed in README.modperl2) to the interface
-provided by this module under mod_perl version 2, please contact me at
-mschout@gkg.net.  The interface for mod_perl version 1 has not changed.
+This module is for mod_perl version 2.  If you are running mod_perl version 1,
+you should be using B<Apache::AuthCookie> instead.
 
-B<Apache::AuthCookie> allows you to intercept a user's first
+B<Apache2::AuthCookie> allows you to intercept a user's first
 unauthenticated access to a protected document. The user will be
 presented with a custom form where they can enter authentication
 credentials. The credentials are posted to the server where AuthCookie
@@ -700,8 +723,7 @@ displaying the user's credentials in the Location field. They don't really
 change AuthCookie's model, but they do add another round-trip request to the
 client.
 
-=for html
-<PRE>
+=for html <PRE>
 
  (-----------------------)     +---------------------------------+
  ( Request a protected   )     | AuthCookie sets custom error    |
@@ -768,12 +790,11 @@ client.
     create must be able to determine if this session_key is valid and
     map it back to the originally authenticated user ID.
 
-=for html
-</PRE>
+=for html </PRE>
 
 =head1 METHODS
 
-C<Apache::AuthCookie> has several methods you should know about.  Here
+C<Apache2::AuthCookie> has several methods you should know about.  Here
 is the documentation for each. =)
 
 =over 4
@@ -801,7 +822,7 @@ return OK on success and HTTP_FORBIDDEN on failure.
 =item * authen_cred()
 
 You must define this method yourself in your subclass of
-C<Apache::AuthCookie>.  Its job is to create the session key that will
+C<Apache2::AuthCookie>.  Its job is to create the session key that will
 be preserved in the user's cookie.  The arguments passed to it are:
 
  sub authen_cred ($$\@) {
@@ -824,7 +845,7 @@ doesn't tamper with the session key.  More info in the Eagle book.
 =item * authen_ses_key()
 
 You must define this method yourself in your subclass of
-Apache::AuthCookie.  Its job is to look at a session key and determine
+Apache2::AuthCookie.  Its job is to look at a session key and determine
 whether it is valid.  If so, it returns the username of the
 authenticated user.
 
@@ -867,7 +888,7 @@ again').
     return($CODE);
   }
           
-  where CODE is a valid code from Apache::Constants
+  where CODE is a valid code from Apache2::Const
 
 =item * login()
 
@@ -887,6 +908,17 @@ specified with the C<PerlSetVar WhatEverLoginScript> configuration
 directive. You can overwrite this method to provide your own
 mechanism.
 
+=item * login_form_status($r)
+
+This method returns the HTTP status code that will be returned with the login
+form response.  The default behaviour is to return HTTP_FORBIDDEN, except for
+some known browsers which ignore HTML content for HTTP_FORBIDDEN responses
+(e.g.: SymbianOS).  You can override this method to return custom codes.
+
+Note that HTTP_FORBIDDEN is the most correct code to return as the given
+request was not authorized to view the requested page.  You should only change
+this if HTTP_FORBIDDEN does not work.
+
 =item * logout()
 
 This is simply a convenience method that unsets the session key for
@@ -925,7 +957,7 @@ only limited set of characters.
 
 =head1 EXAMPLE
 
-For an example of how to use Apache::AuthCookie, you may want to check
+For an example of how to use Apache2::AuthCookie, you may want to check
 out the test suite, which runs AuthCookie through a few of its paces.
 The documents are located in t/eg/, and you may want to peruse
 t/real.t to see the generated httpd.conf file (at the bottom of
@@ -1030,22 +1062,12 @@ Now C<authen_ses_key()> looks up the C<$ses_key> passed to it and
 returns the saved login.  I use Oracle to store the session key and
 retrieve it later, see the ToDo section below for some other ideas.
 
-=head1 KNOWN LIMITATIONS
-
-If the first unauthenticated request is a POST, it will be changed to
-a GET after the user fills out the login forms, and POSTed data will
-be lost.
-
 =head2 TO DO
 
 =over 4
 
 =item *
 
-There ought to be a way to solve the POST problem in the LIMITATIONS
-section.  It involves being able to re-insert the POSTed content into
-the request stream after the user authenticates.
-
 It might be nice if the logout method could accept some parameters
 that could make it easy to redirect the user to another URI, or
 whatever.  I'd have to think about the options needed before I
@@ -1053,13 +1075,7 @@ implement anything, though.
 
 =back
 
-=head1 CVS REVISION
-
-$Id: AuthCookie.pm 226 2008-03-03 17:37:04Z mschout $
-
-=head1 AUTHOR
-
-Michael Schout <mschout@gkg.net>
+=head1 HISTORY
 
 Originally written by Eric Bartley <bartley@purdue.edu>
 
@@ -1076,6 +1092,30 @@ the same terms as Perl itself.
 
 L<perl(1)>, L<mod_perl(1)>, L<Apache(1)>.
 
+=head1 SOURCE
+
+The development version is on github at L<http://github.com/mschout/apache-authcookie>
+and may be cloned from L<git://github.com/mschout/apache-authcookie.git>
+
+=head1 BUGS
+
+Please report any bugs or feature requests to bug-apache-authcookie@rt.cpan.org or through the web interface at:
+ http://rt.cpan.org/Public/Dist/Display.html?Name=Apache-AuthCookie
+
+=head1 AUTHOR
+
+Michael Schout <mschout@cpan.org>
+
+=head1 COPYRIGHT AND LICENSE
+
+This software is copyright (c) 2000 by Ken Williams.
+
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
+
 =cut
 
+__END__
+
+
 # vim: sw=4 ts=4 ai et
